//
//  WizardPageController.swift
//  Hydrometra
//
//  Created by Lukáš Šanda on 11.04.18.
//  Copyright © 2018 Lukáš Šanda. All rights reserved.
//

import UIKit


struct WizardStep {
    let title: String
    let subtitle: String
    let image: UIImage?
}

class WizardPageController: UIPageViewController, UIPageViewControllerDelegate, UIPageViewControllerDataSource {
    
    // MARK: - Proměnné
    
    var pageControl: UIPageControl!
    var btnBack: UIButton!
    var btnNext: UIButton!
    
    let wizardSteps = [
        WizardStep(
            title: NSLocalizedString("WIZARD_WELCOME_TITLE".localized, comment: "Welcome wizard step 1 title"),
            subtitle: NSLocalizedString("WIZARD_WELCOME_SUBTITLE".localized, comment: "Welcome wizard step 1 subtitle"),
            image: #imageLiteral(resourceName: "WizardWelcome")),
        WizardStep(
            title: NSLocalizedString("WIZARD_MONITORING_TITLE".localized, comment: "Welcome wizard step 2 title"),
            subtitle: NSLocalizedString("WIZARD_MONITORING_SUBTITLE".localized, comment: "Welcome wizard step 2 subtitle"),
            image: #imageLiteral(resourceName: "WizardMonitoring")),
        WizardStep(
            title: NSLocalizedString("WIZARD_EDIT_TITLE".localized, comment: "Welcome wizard step 3 title"),
            subtitle: NSLocalizedString("WIZARD_EDIT_SUBTITLE".localized, comment: "Welcome wizard step 3 subtitle"),
            image: #imageLiteral(resourceName: "WizardManage"))
    ]
    
    // MARK: - Funkce
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        if wizardSteps.count > 0 {
            let firstController = getStepController(0)!
            let startingControllers = [firstController]
            self.setViewControllers(startingControllers, direction: .forward, animated: false, completion: nil)
        }
        
        self.dataSource = self
        self.delegate = self
        
        initComponents()
    }
    
    // MARK: Delegované funkce komponenty UIPageControl
    
    func presentationCount(for pageViewController: UIPageViewController) -> Int {
        return 0
    }
    
    func presentationIndex(for pageViewController: UIPageViewController) -> Int {
        return 0
    }
    
    func pageViewController(_ pageViewController: UIPageViewController, didFinishAnimating finished: Bool, previousViewControllers: [UIViewController], transitionCompleted completed: Bool) {
        if !completed { return }
        if let currentViewController = self.viewControllers?.last as? WizardStepController {
            let currentIndex = currentViewController.stepIndex
            
            self.pageControl.currentPage = currentIndex!
        }
    }
    
    // MARK: Funkce datového zdroje pro komponentu UIPageControl
    
    func pageViewController(_ pageViewController: UIPageViewController, viewControllerBefore viewController: UIViewController) -> UIViewController? {
        let stepController = viewController as! WizardStepController
        
        if stepController.stepIndex > 0 {
            return getStepController(stepController.stepIndex - 1)
        }
        
        return nil
    }
    
    func pageViewController(_ pageViewController: UIPageViewController, viewControllerAfter viewController: UIViewController) -> UIViewController? {
        let stepController = viewController as! WizardStepController
        
        if stepController.stepIndex + 1 < wizardSteps.count {
            return getStepController(stepController.stepIndex + 1)
        }
        
        return nil
    }
    
    // MARK: - Privátní funkce
    
    /// Funkce pro konfiguraci komponent Page Control a Button
    private func initComponents() {
        pageControl = UIPageControl()
        pageControl.numberOfPages = wizardSteps.count
        pageControl.currentPage = 0
        pageControl.pageIndicatorTintColor = .secondaryColor
        pageControl.currentPageIndicatorTintColor = .primaryColor
        
        self.view.addSubview(pageControl)
        
        pageControl.snp.makeConstraints {
            $0.centerX.equalToSuperview()
        }
        
        btnBack = UIButton(type: .system)
        btnBack = UIButton()
        btnBack.setTitle("WIZARD_BUTTON_BACK".localized, for: UIControlState())
        btnBack.setTitleColor(.primaryColor, for: UIControlState())
        btnBack.titleLabel?.font = UIFont.systemFont(ofSize: 17)
        btnBack.addTarget(self, action: #selector(buttonBackTouched), for: .touchUpInside)
        
        self.view.addSubview(btnBack)
        
        btnBack.snp.makeConstraints {
            $0.centerY.equalTo(pageControl)
            if #available(iOS 11.0, *) {
                $0.bottom.equalTo(view.safeAreaLayoutGuide).inset(18)
                $0.leading.equalTo(view.safeAreaLayoutGuide).inset(24)
            } else {
                $0.bottom.equalToSuperview().inset(18)
                $0.leading.equalToSuperview().inset(24)
            }
        }
        
        btnNext = UIButton(type: .system)
        btnNext.setTitle("WIZARD_BUTTON_NEXT".localized, for: UIControlState())
        btnNext.setTitleColor(.primaryColor, for: UIControlState())
        btnNext.titleLabel?.font = UIFont.systemFont(ofSize: 17)
        btnNext.addTarget(self, action: #selector(buttonNextTouched), for: .touchUpInside)
        
        self.view.addSubview(btnNext)
        
        btnNext.snp.makeConstraints {
            $0.centerY.equalTo(pageControl)
            if #available(iOS 11.0, *) {
                $0.bottom.equalTo(view.safeAreaLayoutGuide).inset(18)
                $0.trailing.equalTo(view.safeAreaLayoutGuide).inset(24)
            } else {
                $0.bottom.equalToSuperview().inset(18)
                $0.trailing.equalToSuperview().inset(24)
            }
        }
    }
    
    
    /// Funkce pro získání třídy pro následující prvek v komponentě UIPageControl
    ///
    /// - Parameter stepIndex: Index třídy
    /// - Returns: Nalezená třída
    private func getStepController(_ stepIndex: Int) -> WizardStepController? {
        if stepIndex < wizardSteps.count {
            let stepController = storyboard!.instantiateViewController(withIdentifier: "WizardStepViewController") as? WizardStepController
            
            stepController?.titleText = wizardSteps[stepIndex].title
            stepController?.subtitleText = wizardSteps[stepIndex].subtitle
            stepController?.image = wizardSteps[stepIndex].image
            
            stepController?.stepIndex = stepIndex
            return stepController
        } else {
            return nil
        }
    }
    
    /// Funkce pro zachycení události kliknutí na tlačítko "Další"
    ///
    /// - Parameter sender: Objekt, který událost odeslal
    @objc private func buttonNextTouched(sender: UIButton!) {
        if self.pageControl.currentPage < wizardSteps.count - 1 {
            if var currentIndex = (viewControllers?.last as? WizardStepController)?.stepIndex {
                
                currentIndex += 1
                
                if let nextController = getStepController(currentIndex) {
                    self.pageControl.currentPage = currentIndex
                    
                    self.setViewControllers([nextController], direction: .forward, animated: true, completion: nil)
                }
            }
        } else {
            let measureStoryboard: UIStoryboard = .main
            let viewController = measureStoryboard.instantiateViewController(withIdentifier: "LoginViewController")
            self.present(viewController, animated: true, completion: {
                UserDefaults.standard.isWizardDisplayed = true
            })
        }
    }
    
    /// Funkce pro zachycení události kliknutí na tlačítko "Zpět"
    ///
    /// - Parameter sender: Objekt, který událost odeslal
    @objc private func buttonBackTouched(sender: UIButton!) {
        if self.pageControl.currentPage > 0 {
            if var currentIndex = (viewControllers?.last as? WizardStepController)?.stepIndex {
                
                currentIndex -= 1
                
                if let nextController = getStepController(currentIndex) {
                    self.pageControl.currentPage = currentIndex
                    
                    self.setViewControllers([nextController], direction: .reverse, animated: true, completion: nil)
                }
            }
        }
    }
}

