//
//  ConsumptionBarChart.swift
//  Hydrometra
//
//  Created by Lukáš Šanda on 04.02.18.
//  Copyright © 2018 Lukáš Šanda. All rights reserved.
//

import Foundation
import Charts


// MARK: - Pomocná třída pro vytvoření sloupcového grafu
class ConsumptionBarChart {
    
    // MARK: - Proměnné
    
    let senzor: Senzor
    var barChart: BarChartView
    
    init(_ barChart: BarChartView, for senzor: Senzor) {
        self.senzor = senzor
        self.barChart = barChart
    }
    
    // MARK: - Funkce
    
    /// Funkce pro vytvoření sloupcového grafu
    ///
    /// - Parameter type: Období, pro které má být graf vytvořen
    /// - Returns: Vytvořený graf
    func createBarChart(type: TimeUnit) -> BarChartView? {
        var labelDescription = [String]()
        var values = [Double]()
        
        switch type {
        case .day:
            for rawHour in Utils.shared.typeOfCons(type, for: senzor) {
                let hour = rawHour as! NSDictionary
                let rawDate = hour["from"] as! String
                if let hourY = Utils.shared.formatDate(rawDate, format: "H") {
                    labelDescription.append(hourY)
                    values.append(hour["value"] as! Double * 1000)
                }
            }
            
        case .month:
            for rawDay in Utils.shared.typeOfCons(type, for: senzor) {
                let day = rawDay as! NSDictionary
                let rawDate = day["from"] as! String
                if let date = Utils.shared.formatDate(rawDate, format: "d") {
                    labelDescription.append(date)
                    values.append(day["value"] as! Double)
                }
            }
            
        case .year:
            for rawMonth in Utils.shared.typeOfCons(type, for: senzor) {
                let month = rawMonth as! NSDictionary
                let rawMonth = month["from"] as! String
                if let monthX = Utils.shared.formatDate(rawMonth, format: "MMM") {
                    labelDescription.append(monthX)
                    values.append(month["value"] as! Double)
                }
            }
        }
        
        barChart.setBarChartData(xValues: labelDescription, yValues: values, label: "")
        
        barChart.leftAxis.axisMaximum = values.isEmpty ? 1 : values.max()! + (values.max()! / 10)
        barChart.xAxis.labelCount = type == .year ? 12 : Int(Double(values.count) / Double(2))

        barChart.isUserInteractionEnabled = values.sorted(by: { $0 > $1 }).first != 0 ? true : false
        barChartAppereance()

        return barChart
    }
    
    /// Pomocná funkce, která spravuje vizuální stránku grafu
    private func barChartAppereance() {
        barChart.xAxis.granularityEnabled = true
        barChart.xAxis.granularity = 1.0
        barChart.xAxis.labelPosition = .bottom
        barChart.xAxis.drawGridLinesEnabled = false
        
        barChart.leftAxis.drawLabelsEnabled = false
        barChart.leftAxis.axisMinimum = 0
        barChart.leftAxis.drawGridLinesEnabled = false
        
        barChart.rightAxis.drawLabelsEnabled = false
        
        barChart.autoScaleMinMaxEnabled = true
        barChart.legend.enabled = false
        barChart.chartDescription?.enabled = false
        barChart.scaleYEnabled = false
        barChart.doubleTapToZoomEnabled = false
    }
}

// MARK: - Rozšíření, které umožňuje nastavení vlastních popisných hodnot na osu X
extension BarChartView {
    
    private class BarChartFormatter: NSObject, IAxisValueFormatter {
        
        var labels: [String] = []
        
        func stringForValue(_ value: Double, axis: AxisBase?) -> String {
            if Int(value) < labels.count {
                return labels[Int(value)]
            }
            
            return ""
        }
        
        init(labels: [String]) {
            super.init()
            
            self.labels = labels
        }
    }
    
    func setBarChartData(xValues: [String], yValues: [Double], label: String) {
        var dataEntries: [BarChartDataEntry] = []

        for i in 0..<yValues.count {
            let dataEntry = BarChartDataEntry(x: Double(i), y: yValues[i])
            dataEntries.append(dataEntry)
        }
        
        let chartDataSet = BarChartDataSet(values: dataEntries, label: label)
        chartDataSet.colors = [.primaryColor]
        
        let formatter = NumberFormatter()
        formatter.zeroSymbol = ""
        formatter.maximumFractionDigits = 2
        
        chartDataSet.valueFormatter = DefaultValueFormatter(formatter: formatter)
        
        let chartData = BarChartData(dataSet: chartDataSet)
        
        let chartFormatter = BarChartFormatter(labels: xValues)
        let xAxis = XAxis()
        xAxis.valueFormatter = chartFormatter
        self.xAxis.valueFormatter = xAxis.valueFormatter
        
        self.data = chartData
    }
    
}
