//
//  Utils.swift
//  Hydrometra
//
//  Created by Lukáš Šanda on 19.01.18.
//  Copyright © 2018 Lukáš Šanda. All rights reserved.
//

import Foundation

class Utils {
    
    /// Sdílená proměnná
    static let shared: Utils = { return Utils() }()
    
    // MARK: - Pomocná funkce pro naformátování datumu
    /**
     Převedení datumu ve formátu ISO8601 na požadovaný formát
     
     - Parameter string: Řetězec pro formátování
     - Parameter format: Příslušný format: DD.MM.YYYY; DD.MM, YYYY-MM-DD aj.
     
     - Returns: Naformátovaný datum
     */
    func formatDate(_ string: String, format: String = "yyyy-MM-dd'T'HH:mm:ss'Z'", isUTC: Bool = false) -> String? {
        let dateFormatter = DateFormatter()
        
        dateFormatter.locale = Locale(identifier: NSLocale.current.languageCode!) // Nastavení lokalizace v rámci jazyka zařízení
        dateFormatter.timeZone = TimeZone(secondsFromGMT: 0)
        dateFormatter.dateFormat = "yyyy-MM-dd'T'HH:mm:ss'Z'"
        
        if let date = dateFormatter.date(from: string) {
            dateFormatter.dateFormat = format
            dateFormatter.timeZone = .current
            
            let result = dateFormatter.string(from: date)
            return result
        }
        
        return nil
    }
    
    // MARK: - Pomocná funkce pro konvertování stavu vodoměru ve formátu Int na String
    /**
     Pomocná funkce pro konvertování stavu vodoměru ve formátu Int na String
     
     - Parameter state: Stav vodoměru ve formátu Int
     - Returns: Přislušný stav ve formátu String
     */
    func convertStateToString(_ state: Int!) -> String! {
        switch state {
        case 1:
            return "STATE_OPEN".localized
        case 2:
            return "STATE_OPEN_REQUIRED".localized
        case 3:
            return "STATE_CLOSE".localized
        case 4:
            return "STATE_CLOSE_REQUIRED".localized
        default:
            return RuntimeError("Nastala chyba při konvertování.").message
        }
    }
    
    // MARK: - Pomocná funkce vracející aktuální datum a čas
    /// Pomocná funkce vracející aktuální datum a čas
    func getCurrentDateTime() -> String {
        
        let date = Date()
        
        var calender = Calendar.current
        calender.locale = Locale(identifier: NSLocale.current.languageCode!)
        calender.timeZone = TimeZone(identifier: "UTC")!
        let components = calender.dateComponents([.year,.month,.day,.hour,.minute,.second], from: date)
        
        let year = components.year
        let month = components.month
        let day = components.day
        let hour = components.hour
        let minute = components.minute
        let second = components.second
        
        let currentDateTime = String(year!) + "-" + String(month!) + "-" + String(day!) + "T" + String(hour!)  + ":" + String(minute!) + ":" +  String(second!) + "Z"
        
        return currentDateTime
    }
    

    func getDateBasedOnOffset(timeUnit: TimeUnit, _ value: Int, date: Date) -> Date? {
        switch timeUnit {
        case .day:
            return Calendar.current.date(byAdding: .day, value: value, to: date)!
        case .month:
            return Calendar.current.date(byAdding: .month, value: value, to: date)!
        case .year:
            return Calendar.current.date(byAdding: .year, value: value, to: date)!
        }
    }
    
    
    // MARK: Funkce pro získání spotřeby za požadované období
    /**
     Funkce pro získání spotřeby za požadované období
     
     - Parameter type: Období pro výpočet (week/month/year)
     
     - Returns: Požadovaná spotřeba
     */
    func typeOfCons(_ type: TimeUnit, for senzor: Senzor) -> NSArray {
        var typeOfSelectedCons = NSArray()
        
        switch type {
        case .day:
            typeOfSelectedCons = (senzor.consumptionDay)!
            break
        case .month:
            typeOfSelectedCons = (senzor.consumptionMonth)!
            break
        case .year:
            typeOfSelectedCons = (senzor.consumptionYear)!
            break
        }
        
        return typeOfSelectedCons
    }
    
    // MARK: Výpočet spotřeby za parametrem dané obdoví
    /**
     Výpočet spotřeby za parametrem dané obdoví
     
     - Parameter type: Období pro výpočet (week/month/year)
     
     - Returns: Hodnota spotřeby za zadané obdoví
     */
    func calcConsumption(_ type: TimeUnit, for senzor: Senzor) -> Double {
        var consumption = 0.0
        
        for rawVal in Utils.shared.typeOfCons(type, for: senzor) {
            let val = rawVal as! NSDictionary
            let value = val["value"] as! Double
            consumption += value
        }
        
        return consumption
    }
    
    func resetApplication() {
        VodomerData.shared.tempSenzor = nil
        
        UserDefaults.standard.userRole = ""
        UserDefaults.standard.userLoginToken = ""
    }
}
