//
//  ConsumptionAPI.swift
//  Hydrometra
//
//  Created by Lukáš Šanda on 19.01.18.
//  Copyright © 2018 Lukáš Šanda. All rights reserved.
//

import Foundation
import Alamofire


class ConsumptionAPI: AlamofireManager {
    
    // MARK: Sdílená proměnná
    static let shared: ConsumptionAPI = { return ConsumptionAPI() }()
    
    /// Asynchronní požadavek na databázi pro získání celkové spotřeby vodoměru
    ///
    /// - Parameters:
    ///   - id: Identifikační číslo senzoru
    ///   - completionHandler: Výstupní parametr pro zachycení asynchronního výstupu
    func requestTotalConsumption(id: Int, completionHandler: @escaping (_ responseObject: Double, _ error: Error?) -> ()) {
        
        let url = "\(baseURL)/api-m/Sensor/\(id.description)/consumption/total"
        
        headers.updateValue(UserDefaults.standard.userLoginToken, forKey: "Authorization")
        
        Manager.request(url, headers: headers).responseJSON { (response) in
            switch response.result {
            case .success:
                if let json = response.result.value as? NSDictionary {
                    
                    if let result  = json["consTotal"] as? Double {
                        completionHandler(result, nil)
                    } else {
                        completionHandler(Double.nan, RuntimeError("Hodnota požadavku není konzistentní."))
                    }
                } else {
                    completionHandler(Double.nan, RuntimeError("Chyba při dynamickém přetypování."))
                }
            case .failure(let error):
                completionHandler(Double.nan, error)
            }
        }
    }
    
    
    /// Asynchronní funkce pro získání spotřeby za dané období
    ///
    /// - Parameters:
    ///   - id: Identifikační číslo senzoru
    ///   - offset: Hodnota pro zajištění aktuálního, předchozího či následujícího data
    ///   - timeUnit: Období spotřeby
    ///   - date: Datum od kterého má být offset počítán
    ///   - completionHandler: Výstupní parametr pro zachycení asynchronního výstupu
    func requestConsumption(id: Int, offset: Int, timeUnit: TimeUnit, date: String? = nil, completionHandler: @escaping (_ responseObject: Senzor?, _ error: Error?) -> ()) {
        let url = "\(baseURL)/api-m/Sensor/\(id)/consumption"
        
        let date: String = date ?? Utils.shared.formatDate(Utils.shared.getCurrentDateTime(), format: "YYYY-MM-dd'T'HH:mm:ss'Z'")!
        
        let parameters: Parameters = [
            "timeUnit": timeUnit.rawValue,
            "offset": offset,
            "dateTime": date
        ]
        
        headers.updateValue(UserDefaults.standard.userLoginToken, forKey: "Authorization")
        
        
        Manager.request(url, method: .post, parameters: parameters, encoding: JSONEncoding.default, headers: headers).responseJSON { (response) in
            switch response.result {
            case .success:
                if let json = response.result.value as? NSDictionary {
                    if let senzor = VodomerData.shared.senzors.first(where: {$0.idHw == id}) {
                        switch timeUnit {
                        case .day:
                            if let jsonArray = json["data"] as? NSArray {
                                
                                senzor.consumptionDay = jsonArray
                                completionHandler(senzor, nil)
                            } else {
                                completionHandler(nil, RuntimeError("Chyba při dynamickém přetypování."))
                            }
                        case .month:
                            if let jsonArray = json["data"] as? NSArray {
                                senzor.consumptionMonth = jsonArray
                                completionHandler(senzor, nil)
                            } else {
                                completionHandler(nil, RuntimeError("Chyba při dynamickém přetypování."))
                            }
                        case .year:
                            if let jsonArray = json["data"] as? NSArray {
                                senzor.consumptionYear = jsonArray
                                completionHandler(senzor, nil)
                            } else {
                                completionHandler(nil, RuntimeError("Chyba při dynamickém přetypování."))
                            }
                        }
                        
                    } else {
                        completionHandler(nil, RuntimeError("Chyba při dynamickém přetypování."))
                    }
                } else {
                    completionHandler(nil,RuntimeError("Senzor nenalezen."))
                }
            case .failure(let error):
                completionHandler(nil, error)
            }
        }
    }
    
    
    /// Asynchronní funkce pro změnu celkové spotřeby senzoru
    ///
    /// - Parameters:
    ///   - id: Indentifikační číslo senzoru
    ///   - consumption: Nová spotřeba
    ///   - completionHandler: Výstupní parametr pro zachycení asynchronního výstupu
    func adjustTotalConsumption(id: Int, consumption: Double, completionHandler: @escaping (_ responseObject: Senzor?, _ error: Error?) -> ()) {
        let url = "\(baseURL)/api-m/Sensor/\(id.description)/adjustTotalConsumption"
        
        let parameters: Parameters = [
            "consumption": consumption
        ]
        
        headers.updateValue(UserDefaults.standard.userLoginToken, forKey: "Authorization")
        
        Manager.request(url, method: .post, parameters: parameters, encoding: JSONEncoding.default, headers: headers).responseString { (response) in
            switch response.result {
            case .success:
                if let senzor = VodomerData.shared.senzors.first(where: {$0.idHw == id}) {
                    senzor.consumptionTotal = parameters["consumption"] as? Double
                    
                    completionHandler(senzor, nil)
                } else {
                    completionHandler(nil, RuntimeError("Senzor nenalezen."))
                }
            case .failure(let error):
                completionHandler(nil, error)
            }
        }
    }    
}
