//
//  RESTAPI.swift
//  Hydrometra
//
//  Created by Lukáš Šanda on 21.11.17.
//  Copyright © 2017 Lukáš Šanda. All rights reserved.
//

import Foundation
import Alamofire


// MARK: - Třída zajíšťující komunikaci s REST API
/// Třída obsahující veškeré funkce spojené s requesty na databázi pro obsluhu komunikace pomocí REST API.
class SenzorAPI: AlamofireManager {
    
    // MARK: Sdílená proměnná
    static let shared: SenzorAPI = { return SenzorAPI() }()
    
    // MARK: - Funkce
    
    // MARK: Request na daný vodoměr
    /**
     Asynchronní požadavek na databázi na daný vodoměr pro získání doplňujících informací
     
     - Parameter id: ID požadovaného vodoměru
     - Parameter completionHandler: zajištění dokončení v hlavním vlákně
     
     - Returns: Úspěch – daný vodoměr, Neúspěch – Error
     */
    func requestSenzor(_ id: Int, completionHandler: @escaping (_ responseObject: Senzor?, _ error: Error?) -> ()) {
        let url = "\(baseURL)/api-m/Sensor/\(id.description)"
        
        headers.updateValue(UserDefaults.standard.userLoginToken, forKey: "Authorization")
        
        Manager.request(url, headers: headers).responseJSON { (response) in
            switch response.result {
            case .success:
                if let json = response.result.value as? NSDictionary {
                    if let senzor = VodomerData.shared.senzors.first(where: {$0.idHw == id}) {
                        senzor.countStop = json["countStop"] as? Int ?? 0
                        senzor.countStopNight = json["countStopNight"] as? Int ?? 0
                        senzor.nightStartHour = json["nightStartHour"] as? Int ?? 0
                        senzor.nightEndHour = json["nightEndHour"] as? Int ?? 0
                        senzor.limitDay = json["limitDay"] as? Double ?? 0.0
                        senzor.limitMonth = json["limitMonth"] as? Double ?? 0.0
                        senzor.state = json["state"] as? Int ?? 0
                        
                        completionHandler(senzor, nil)
                    } else {
                        completionHandler(nil, RuntimeError("Senzor nenalezen."))
                    }
                } else {
                    completionHandler(nil, RuntimeError("Chyba při dynamickém přetypování."))
                }
            case .failure(let error):
                completionHandler(nil, error)
            }
        }
    }
    
    // MARK: Request na všechny vodoměry
    /**
     Asynchronní požadavek na databázi pro získání základních informací o vodoměrech
     
     - Parameter completionHandler: zajištění dokončení v hlavním vlákně
     
     - Returns: Úspěch – vodoměry, Neúspěch – Error
     */
    func requestSenzors(completionHandler: @escaping (_ responseObject: [Senzor]?, _ error: Error?) -> ()) {
        let url = "\(baseURL)/api-m/Sensor"
        
        headers.updateValue(UserDefaults.standard.userLoginToken, forKey: "Authorization")
        
        Manager.request(url, headers: headers).responseJSON { (response) in
            switch response.result {
            case .success:
                if let jsonArray = response.result.value as? NSArray {
                    
                    var senzors = [Senzor]()
                    for json in jsonArray {
                        senzors.append(Senzor(data: json as! [String : Any]))
                    }
                    //VodomerData.shared.senzors = senzors
                    completionHandler(senzors, nil)
                } else {
                    completionHandler(nil, RuntimeError("Chyba při dynamickém přetypování."))
                }
            case .failure(let error):
                completionHandler(nil, error)
            }
        }
    }
    
    
    /// Asychronní funkce pro změnu stavu ventilu
    ///
    /// - Parameters:
    ///   - id: Identifiační číslo senzoru
    ///   - state: Požadovaný stav
    ///   - completionHandler: Výstupní parametr pro zachycení asynchronního výstupu
    func putValveState(id: Int, state: Int, completionHandler: @escaping (_ responseObject: Senzor?, _ error: Error?) -> ()) {
        
        let stateString = state == 4 ? "CloseRequired" : "OpenRequired"
        let url = "http://petrkoula-001-site1.ctempurl.com/api-m/Valve/\(id)/state?name=\(stateString)"
        
        headers.updateValue(UserDefaults.standard.userLoginToken, forKey: "Authorization")
        
        
        Alamofire.request(url, method: .put, headers: headers).responseString { (response) in
            switch response.result {
            case .success:
                if let senzor = VodomerData.shared.senzors.first(where: {$0.idHw == id}) {
                    
                    senzor.state = state
                    senzor.stateModifierUserId = 1
                    senzor.stateModifiedDate = Utils.shared.getCurrentDateTime()
                    
                    completionHandler(senzor, nil)
                } else {
                    completionHandler(nil, RuntimeError("Senzor nenalezen."))
                }
            case .failure(let error):
                completionHandler(nil, error)
            }
        }
    }
    
    
    /// Asynchronní funkce pro aktualizaci senzoru
    ///
    /// - Parameters:
    ///   - id: Identifikační číslo senzoru
    ///   - parameters: Nové hodnoty
    ///   - completionHandler: Výstupní parametr pro zachycení asynchronního výstupu
    func updateSenzor(id: Int, parameters: Parameters, completionHandler: @escaping (_ responseObject: Senzor?, _ error: Error?) -> ()) {
        
        let url = "\(baseURL)/api-m/Sensor/\(id.description)/update"
    
        headers.updateValue(UserDefaults.standard.userLoginToken, forKey: "Authorization")
        
        Manager.request(url, method: .post, parameters: parameters, encoding: JSONEncoding.default, headers: headers).responseString { (response) in
            switch response.result {
            case .success:
                
                if let senzor = VodomerData.shared.senzors.first(where: {$0.idHw == id}) {
                    
                    senzor.nameHw = parameters["nameHw"] as? String ?? ""
                    senzor.pricePerM3 = Double(parameters["pricePerM3"] as! String) ?? 0.0
                    senzor.implPerLit = Int(parameters["implPerLit"] as! Int)
                    senzor.countStop = Int(parameters["countStop"] as! String) ?? 0
                    senzor.countStopNight = Int(parameters["countStopNight"] as! String) ?? 0
                    senzor.nightStartHour = Int(parameters["nightStartHour"]as! String) ?? 0
                    senzor.nightStartMinute = Int(parameters["nightStartMinute"] as! String) ?? 0
                    senzor.nightEndHour = Int(parameters["nightEndHour"] as! String) ?? 0
                    senzor.nightEndMinute = Int(parameters["nightEndMinute"] as! String) ?? 0
                    senzor.limitDay = Double(parameters["limitDay"] as! String) ?? Double.nan
                    senzor.limitMonth = Double(parameters["limitMonth"] as! String) ?? Double.nan
                    
                    completionHandler(senzor, nil)
                } else {
                    completionHandler(nil, RuntimeError("Senzor nenalezen."))
                }
            case .failure(let error):
                completionHandler(nil, error)
            }
        }
    }
}
